"""
navigation.py defines classes that allow the user to navigate between and pick formulas.
"""
from formula_prompt.shared import MAX_ENTRY_ATTEMPTS, UserInputError


class _Element:
    """Element that can be displayed in the Navigation. Sub-classes include Folder and Formula"""

    def __init__(self, name):
        """
        :param name: Name to display in the navigation
        """
        self.name = name

    def run(self):
        """
        Called when the element is selected during navigation
        """
        raise NotImplementedError


class Formula(_Element):
    def __init__(self, func, inputs, name):
        super(Formula, self).__init__(name)
        self.func = func
        self.inputs = inputs

    def run(self):
        while True:
            inputs = []

            # For each required input, read the input and add it the list
            try:
                for input_description in self.inputs:
                    inputs.append(input_description.read())
            # If the user fails to enter an input, cancel the formula
            except UserInputError:
                break

            # Call the formula with the inputs
            ans = self.func(*inputs)

            # Print the results
            if ans is not None:
                print(f"{self.name}:\n{ans}")

            selection = input("\nEnter to run again or 1 to return...\n>>> ")
            if selection == "1":
                break


class _Folder:
    """
    A folder or directory that can store other folders or formulas.
    """

    def __init__(self, name, is_root=False):
        """
        :param name: Folder name
        :param is_root: Specifies if this is the root folder
        """
        self.name = name
        self.contents = set()  # Contents of the folder, starts empty
        self.is_root = is_root

    def pick_element(self):
        # Sorted the contents by name for easy navigation
        contents = sorted(list(self.contents), key=lambda x: x.name)

        if self.is_root:
            contents.append(_LeaveFolder("quit"))  # Add the option to quit the root folder (ends program)
        else:
            contents.insert(0, _LeaveFolder("go back"))  # Add option to go up one folder

        # Print them contents of the folder to the user
        for i, element in enumerate(contents):
            print(f"{i}:\t{element.name}")

        # Let the user pick a number representing the desired element
        for _ in range(MAX_ENTRY_ATTEMPTS):
            selection = input("Pick a formula:\n>>> ")

            try:
                return contents[int(selection)]
            except ValueError:
                print("Invalid input. Try again.")
            except IndexError:
                print("Invalid input. Try again.")
        raise UserInputError

    def run(self):
        while True:
            # Let the user pick an element and then run that element
            should_leave = self.pick_element().run()

            # If we should leave break (which returns to parent folder)
            if should_leave:
                break

    def add_formula(self, formula: Formula, path=None, depth=0):
        """
        Add a formula to the folder. Gets called recursiely if the formula lives in a nested folder.

        :param formula: The Formula to add
        :param path: A list of the names of all the folders and the formula
        :param depth: Current position in the list (how deep we are in the nested folders)
        """
        # If path isn't set, we create it by spliting the name at the dots ('.')
        if path is None:
            path = formula.name.split(".")

        # If we're at the end of the path (no more nested folders) we add the formula the current folder (self)
        if depth == len(path) - 1:
            self.contents.add(formula)
            return

        # Otherwise we need to go into the nested folder
        # The folders name is the formula name up the current folder
        folder_name = ".".join(path[:depth + 1])
        # We check if the folder already exists
        for element in self.contents:
            if isinstance(element, _Folder) and element.name == folder_name:
                # If it does, add the formula to that folder (recursive call)
                element.add_formula(formula, path, depth + 1)
                return

        # If the folder doesn't exist, we create it
        new_folder = _Folder(folder_name)
        # And add the formula to it (recursive call)
        new_folder.add_formula(formula, path, depth + 1)
        # And then add the folder to the current folder
        self.contents.add(new_folder)


class _LeaveFolder(_Element):
    """
    A simple element that will return True when run indicating the callee should exit its call loop.
    """

    def __init__(self, name):
        super(_LeaveFolder, self).__init__(name)

    def run(self):
        return True


# Initialize a root folder that one can add formulas to (via @register_formula decorator)
NAVIGATION_ROOT = _Folder(None, is_root=True)



from formula_prompt.shared import MAX_ENTRY_ATTEMPTS, UserInputError


class Input:
    """
    A parent class that can be extended to allow for different types of inputs to a formula.

    Important functions:

    read() -- Called by the program to retrieve the input value from the user.

    process() -- Function to be overridden by subclasses. Should read from input() and return
            the parsed value that will be passed on to the formula.
    """

    def __init__(self, name, optional=False):
        """Initialize the instance.

        Arguments:
            name: The name of the desired input, will be printed to the user before requesting
                the input.
            optional: Whether this input is required. This parameter should be handled by subclasses
                properly.
        """
        self.name = name
        self.optional = optional

    def read(self):
        # Print "Input <name>: " or "Input data: " if name isn't defined.
        print("Input {}:".format(self.name if self.name is not None else "data"))
        return self.process()

    def process(self):
        raise NotImplemented("Please use a subclass of Input such as NumInput or ListInput")


class NumInput(Input):
    """Input that accepts a number from the user."""

    def __init__(self, name="number", require_int=False, **kwargs):
        super(NumInput, self).__init__(name=name, **kwargs)
        self.require_int = require_int

    def process(self):
        for _ in range(MAX_ENTRY_ATTEMPTS):
            try:
                i = input(">>> ")
                if self.require_int:
                    return int(i)
                else:
                    return float(i)
            except ValueError as e:
                if self.optional and i == "":
                    return None
                print("Invalid number. Try again.")
        raise UserInputError


class IntInput(NumInput):
    """Input that accepts an integer from the user."""

    def __init__(self, name="integer", **kwargs):
        super(IntInput, self).__init__(name, require_int=True, **kwargs)


class ListInput(Input):
    """Input that accepts a list of numbers from the user."""

    def __init__(self, name="list", **kwargs):
        super(ListInput, self).__init__(name=name, **kwargs)

    def process(self):
        values = []
        consecutive_failures = 0
        while True:
            try:
                i = input(">>> ")
                values.append(float(i))
                consecutive_failures = 0
            except ValueError:
                if (values or self.optional) and i == "":
                    return values
                consecutive_failures += 1
                if consecutive_failures == MAX_ENTRY_ATTEMPTS:
                    raise UserInputError
                print("Invalid number. Try again.")


"""
Library that allows registering formulas and then using them in your command prompt.

Important functions:

register_formula() -- Should be used as a function decorator to register
formulas into this library.

launch_prompt() -- Starts the prompt using the registered formulas.
"""
import functools
import math

from formula_prompt.inputs import *
from formula_prompt.navigation import NAVIGATION_ROOT, Formula

_DEFAULT_NUMBER_OF_DECIMALS = 4


def register_formula(func_inputs, decimal_places=_DEFAULT_NUMBER_OF_DECIMALS, name=None):
    """
    Function decorator that adds a formula to the list of registered formulas

    :param func_inputs: Element of type <Input> or list of <Input> elements representing
    the inputs that should be passed to the formula
    :param decimal_places: Number of decimal places to round your answer to before printing
    :param name: A name for the function. If the name contains '.', this will be considered as a folder.
    """
    # If only one argument is passed, wrap it by a tuple
    if isinstance(func_inputs, Input):
        func_inputs = (func_inputs,)

    # Define the decorator
    def decorator(func):
        @functools.wraps(func)  # Make func.__name__ keeps its value
        # Define function to be called when func is called
        def inner_function(*args, **kwargs):
            result = func(*args, **kwargs)  # Call the function

            # Round the result
            if decimal_places is not None and isinstance(result, float) and not math.isnan(result):
                multiplier = 10 ** decimal_places
                result = round(result * multiplier) / multiplier

            return result

        # Register the formula in the root folder (the folder will handle placing it in the right location)
        NAVIGATION_ROOT.add_formula(Formula(inner_function, func_inputs, name if name is not None else func.__name__))
        # Return the wrapped function
        return inner_function

    return decorator


def launch_prompt():
    """
    Launches the prompt at the navigation root folder.
    """
    NAVIGATION_ROOT.run()
